/*
 * this is where all of the actual installation work is performed;
 * it is presented as a (mostly) state-less object to the GUI code
 * (it does hold the superuser password as state)
 *
 * FIXME: i would really like to have a more descriptive name for this class
 */


/*
 *
 */
package installer.common;


/*
 *
 */
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.StreamTokenizer;
import java.net.URL;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.UUID;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


/*
 * FIXME: probably should rename this NativeController, and rename
 *        the specific controllers LinuxNativeController, WindowsNativeController, etc.
 */
public abstract class Controller
{
    /*
     * essentially a struct encapsulating relevant data about an installation
     * of WFDB-SWIG-Matlab
     *
     * note that wsmVersion is an immutable List
     */
    public static class WSMInstallationData
    {
        protected List<Integer> wsmVersion;
        protected String matlabVersion;
        protected String matlabPath;
        protected UUID uuid;

        public WSMInstallationData(List<Integer> wsmVersion,
                                   String matlabVersion,
                                   String matlabPath,
                                   UUID uuid)
        {
            this.wsmVersion = Collections.unmodifiableList(wsmVersion);
            this.matlabVersion = matlabVersion;
            this.matlabPath = matlabPath;
            this.uuid = uuid;
        }

        public List<Integer> getWSMVersion() {
            return wsmVersion;
        }

        public String getMatlabVersion() {
            return matlabVersion;
        }

        public String getMatlabPath() {
            return matlabPath;
        }

        public UUID getUUID() {
            return uuid;
        }
    }


    /*
     * essentially a struct, to store both the password text
     * and the "method," e.g. whether su or sudo is used.
     *
     * FIXME: the "method" is system-dependent, but it would
     *        be much more complicated to generalize this
     *
     * FIXME: make this immutable?
     */
    public static class Password
    {
        //
        public enum Method
        {
            SU, SUDO;
        }

        // ownership of passwordText is taken
        public Password(final char[] passwordText, Method authenticationMethod)
        {
            this.passwordText = passwordText;
            this.authenticationMethod = authenticationMethod;

            // safe destructor
            Runtime.getRuntime().addShutdownHook(new Thread()
                {
                    public void run()
                    {
                        if (passwordText != null)
                        {
                            Arrays.fill(passwordText, (char)0);
                        }
                    }
                });
        }

        //
        public final char[] passwordText;
        public final Method authenticationMethod;
    }


    /*
     * can expect to receive the lines (of, e.g., a file) in order
     *
     * FIXME: move this outside
     */
    public static interface LineFilter
    {
        /*
         * return true to *keep* the line;
         * return false to filter it out
         */
        public boolean matches(String line);
    }


    /*
     * FIXME: move this outside
     */
    public static class MatlabPathFileWSMLineFilter implements LineFilter
    {
        final private Pattern commentLinePattern = Pattern.compile("\\s*#(.*)");
        final private Pattern wsmCommentPattern = Pattern.compile(pathFileCommentRegex);

        private boolean prevLineWasWSMComment;
        private String prevLineWSMPath;

        public MatlabPathFileWSMLineFilter()
        {
            this.prevLineWasWSMComment = false;
            this.prevLineWSMPath = null;
        }

        public boolean matches(String line)
        {
            Matcher matcher;
            Logger logger;


            logger = Logger.getLogger("controller");

            if (this.prevLineWasWSMComment)
            {
                //
                // the previous line was a WSM comment; we expect this line to be a path
                // added by the WSM install., matching that of the comment; if it is not,
                // then something weird has happened (like the user manually edited the
                // file)
                //

                //
                if (this.prevLineWSMPath.equals(line))
                {
                    this.prevLineWasWSMComment = false;
                    this.prevLineWSMPath = null;
                    return false;
                }

                // the line we expected has disappeared; treat it as a line NOT preceded
                // by a WSM comment
                // FIXME: it's a bit hackish to reuse matches() like this, but it works well
                this.prevLineWasWSMComment = false;
                this.prevLineWSMPath = null;
                return matches(line);
            }
            else
            {
                String wsmVersionString;
                String uuidString;
                String pathString;
                List<Integer> wsmVersion;
                UUID uuid;

                // pass through all non-comment lines
                matcher = commentLinePattern.matcher(line);
                if (!matcher.matches())
                {
                    return true;
                }

                // does the comment text not match that as is generated by WSM?
                matcher = wsmCommentPattern.matcher(matcher.group(1));
                if (!matcher.matches())
                {
                    return true;
                }

                //
                wsmVersionString = matcher.group(1);
                uuidString = matcher.group(2);
                pathString = matcher.group(3);
                try
                {
                    wsmVersion = convertStringToVersion(wsmVersionString);
                    uuid = UUID.fromString(uuidString);
                }
                catch (IllegalArgumentException ex)
                {
                    // N.B.: convertStringToVersion() throws NumberFormatException;
                    //       UUID.fromString() throws IllegalArgumentException;
                    //       NumberFormatException is a subclass of IllegalArgumentException
                    
                    logger.info("Strange line encountered while filtering MATLAB *path.txt file: "
                                + "it looks like it was created by the WFDB-SWIG-MATLAB installer but "
                                + "either the version (" + wsmVersionString + ") is misformatted or "
                                + "the UUID (" + uuidString + ") is misformatted.\n");

                    // pass through apparently WSM-commented lines with bad version strings
                    // or bad UUID's
                    return true;
                }

                //
                // at this point, we know that the line was generated by a WSM installation
                // (or is identical to such)
                //

                // set these for use next time matches() is called
                this.prevLineWSMPath = pathString;
                this.prevLineWasWSMComment = true;

                //
                return false;
            }
        }
    }


    /*
     * FIXME: this should be a property from the JNLP file
     */
    static protected final List<Integer> currentVersion = convertStringToVersion("0.0.1");


    /*
     * these describe the comment on the line we add to MATLAB's classpath.txt, and
     * librarypath.txt; it is very important that the regex. match the text (and not
     * much more, hopefully.)
     *
     * FIXME: improve this pattern!
     */
    static final protected String pathFileCommentRegex = " DO NOT EDIT: WSM v\\. (\\S*) (\\S*) (.*)";
    static final protected String pathFileCommentText  = " DO NOT EDIT: WSM v. %s %s %s";


    /*
     *
     */
    static final protected String matlabCodeListFileName = "matlab-files.txt";


    /*
     *
     */
    protected final Logger logger;


    /*
     *
     */
    private final UUID currentUUID;


    /*
     *
     */
    private InstallerGUI gui;


    /*
     *
     */
    private Password superuserPassword;


    /*
     *
     */
    private boolean tempConfigDirHasBeenFinalized;


    /*
     * meant to be accessed directly only by getUserHomeDirectory()
     */
    private File userHomeDirectory;
    private boolean userHomeDirectoryGotten;


    /*
     * a set of the archs. for which native libs. have already
     * been installed
     */
    private Set<String> installedArchs;


    /*
     *
     */
    protected Controller()
    {
        this.logger = Logger.getLogger("controller");
        this.currentUUID = UUID.randomUUID();
        this.gui = null;
        this.superuserPassword = null;
        this.tempConfigDirHasBeenFinalized = false;
        this.userHomeDirectory = null;
        this.userHomeDirectoryGotten = false;
        this.installedArchs = new HashSet<String>();
    }


    /*
     * versionString should consist of (non-negative) integers separated by periods,
     * e.g. "0.1.3" or "1.5" or simply "1"
     */
    private static List<Integer> convertStringToVersion(String versionString)
        throws NumberFormatException
    {
        String[] subVersions;
        List<Integer> version;


        subVersions = versionString.split("\\.");
        version = new LinkedList<Integer>();
        for (String sv : subVersions)
        {
            version.add(Integer.valueOf(sv));
        }

        return version;
    }


    /*
     * version is assumed to be non-null and non-empty
     *
     * FIXME: this could likely be optimized
     */
    public static String convertVersionToString(List<Integer> version)
    {
        String versionString;
        Iterator<Integer> it;


        it = version.iterator();
        versionString = String.valueOf(it.next());
        while (it.hasNext())
        {
            versionString += "." + it.next();
        }

        return versionString;
    }


    /*
     * system-dependent; should attempt to run as superuser
     * if necessary
     */
    protected abstract void copyFile(File sourceFile,
                                     File destinationFile)
        throws IOException,
               NoSuperuserPasswordException,
               ProcessExecutionException;


    /*
     * this *must* not perform any operations that require superuser access
     * (unless, of course, the installer is already being run by the superuser.)
     *
     * archs is the set of MATLAB JVM archs. to be supported -- it should be non-null
     *
     * FIXME: safest to use getActiveConfigDirectory()
     */
    public void copyFilesFromPackageToHost()
        throws IOException,
               ProcessExecutionException,
               ResourceNotFoundException
    {
        File tempConfigDir;
        File finalConfigDir;
        File wsmCodeDir;
        String someArch; // any arch that is used by some jvm


        //
        tempConfigDir = getTempConfigDirectory();
        finalConfigDir = getConfigDirectory();
        someArch = null;

        // remove the old ~/.wfdb-swig-matlab-tmp if it exists
        if (tempConfigDir.exists())
        {
            removeDirectory(tempConfigDir);
        }

        // mkdir ~/.wfdb-swig-matlab-tmp
        makeDirectory(tempConfigDir);

        // make the dir. ~/.wfdb-swig-matlab-tmp/X.Y.Z for the specific current WSM version X.Y.Z
        wsmCodeDir = new File(tempConfigDir, convertVersionToString(currentVersion));
        makeDirectory(wsmCodeDir);

        // mkdir ~/.wfdb-swig-matlab-tmp/X.Y.Z/util
        // FIXME: i don't like that this is hard-coded; we only happen to know that
        //        some of the matlab code files need to go into the util/ subdir.
        makeDirectory(new File(wsmCodeDir, "util"));

        // unpack matlab code files (*.m) into ~/.wfdb-swig-matlab-tmp/X.Y.Z
        for (String codeFileName : getWSMMatlabCodeFileNames())
        {
            copyResourceToLocalFile(codeFileName,
                                    new File(wsmCodeDir, localizeRelativePath(codeFileName)));
        }

        // copy wsm-classes.jar into ~/.wfdb-swig-matlab-tmp/X.Y.Z
        copyResourceToLocalFile("wsm-classes.jar", new File(wsmCodeDir, "wsm-classes.jar"));

        // copy wfdb.jar into ~/.wfdb-swig-matlab-tmp/X.Y.Z
        copyResourceToLocalFile("wfdb.jar", new File(wsmCodeDir, "wfdb.jar"));
    }


    /*
     *
     */
    public void copyNativeLibsFromPackageToHost(Set<String> archs)
        throws IOException,
               ProcessExecutionException,
               ResourceNotFoundException
    {
        File wsmCodeDir;
        String someArch;


        //
        wsmCodeDir = new File(getActiveConfigDirectory(),
                              convertVersionToString(currentVersion));
        someArch = null;
        
        //
        for (String arch : archs)
        {
            File libDir;

            //
            if (this.installedArchs.contains(arch))
            {
                continue;
            }

            //
            libDir = new File(wsmCodeDir, "lib-" + arch);

            // mkdir ~/.wfdb-swig-matlab(-tmp)/X.Y.Z/lib-<arch>
            makeDirectory(libDir);
            
            // unpack native binary libs into ~/.wfdb-swig-matlab(-tmp)/X.Y.Z/lib-<arch>,
            // stripping away the -<arch> suffix from the file names
            for (String libFileName : getNativeBinaryFileNames(arch))
            {
                if (!libFileName.endsWith("-" + arch))
                {
                    // FIXME: shouldn't have to use Logger.getLogger() here
                    this.logger.severe("Error: Erroneously requested lib. name: "
                                       + libFileName
                                       + "; this is a bug!\n");
                    continue;
                }
                else
                {
                    String strippedName =
                        libFileName.substring(0, libFileName.length() - (arch.length() + 1));
                    
                    copyResourceToLocalFile(libFileName,
                                            new File(libDir, localizeRelativePath(strippedName)));
                }
            }
            
            someArch = arch;
            this.installedArchs.add(arch);
        }

        // copy files that are specific to the native platform,
        // from package to host; this is different from the native
        // libs. copied above, because the method of copy and setup
        // is not system-independent
        copyNativeSpecificFilesFromPackageToHost(someArch);
    }        


    /*
     * by default, do nothing
     * FIXME: make this abstract
     */
    protected void copyNativeSpecificFilesFromPackageToHost(String arch)
        throws IOException,
               ProcessExecutionException,
               ResourceNotFoundException
    {
    }


    /*
     *
     */
    protected void copyResourceToLocalFile(String resourceName,
                                           File localFile)
        throws IOException,
               ResourceNotFoundException
    {
        ClassLoader classLoader;
        URL resourceURL;


        //
        classLoader = Thread.currentThread().getContextClassLoader();
        resourceURL = classLoader.getResource(resourceName);
        if (resourceURL == null)
        {
            throw new ResourceNotFoundException(resourceName);
        }

        //
        copyStreamToFile(resourceURL.openStream(), localFile);

        //
        this.logger.fine("Copied resource " + resourceName + " into file "
                         + localFile.getAbsolutePath() + "\n");
    }


    /*
     * if the file exists, remove it;
     * then copy from the stream, into the file, until the end of
     * the stream is reached.
     *
     * FIXME: i don't like having a fixed buffer size, but it's
     *        not really a big deal (but if anything, this buf
     *        should really not be realloc'd every call!)
     *
     * FIXME: add fsync()'s!
     */
    private static void copyStreamToFile(InputStream inputStream,
                                         File outputFile)
        throws IOException
    {
        FileOutputStream fileOutputStream;
        byte[] copyBuf;
        int n;


        //
        copyBuf = new byte[4096];
        fileOutputStream = new FileOutputStream(outputFile);

        //
        try
        {
            while ((n = inputStream.read(copyBuf)) >= 0)
            {
                fileOutputStream.write(copyBuf, 0, n);
            }
        }
        finally
        {
            // FIXME: should we try to remove the new file too, if it wasn't finished?
            fileOutputStream.close();
        }
    }


    /*
     * essentially a wrapper around File.delete() that throws
     * an IOException upon failure. note that fileToDelete may
     * be a directory, but if so, it only makes sense if it's
     * an empty directory (it will certainly fail otherwise.)
     */
    private static void deleteFile(File fileToDelete)
        throws IOException
    {
        Logger logger = Logger.getLogger("controller");   // FIXME: a bit improper


        if (!fileToDelete.delete())
        {
            if (fileToDelete.isDirectory())
            {
                throw new IOException("Unable to delete directory: "
                                      + fileToDelete.getAbsolutePath()
                                      + " (maybe not empty?)");
            }
            else
            {
                throw new IOException("Unable to delete file: " + fileToDelete.getAbsolutePath());
            }
        }
        else
        {
            logger.finer("Deleted " + fileToDelete.getAbsolutePath() + "\n");
        }
    }


    /*
     *
     */
    protected void ensureSuperuserPasswordAvailable()
        throws NoSuperuserPasswordException
    {
        if (this.superuserPassword == null)
        {
            while (true)
            {
                this.superuserPassword = gui.showRootPasswordPrompt();
                if (this.superuserPassword == null)
                {
                    throw new NoSuperuserPasswordException();
                }
                else if (!verifySuperuserPassword(this.superuserPassword))
                {
                    gui.showRootPasswordIncorrectDialog();
                    this.superuserPassword = null;   // just to be safe
                    continue;
                }
                else
                {
                    break;
                }
            }
        }
    }


    /*
     * escape a String containing an absolute path, for use in a matlab *path.txt
     * file (classpath.txt or librarypath.txt); at the moment, the only concrete
     * thing this means is replacing "$" with "$$".
     */
    private static String escapeAbsolutePathForPathFile(String pathString)
    {
        return pathString.replaceAll("\\$", "\\$\\$");
    }


    /*
     * take a string and make it safe for embedding in single quotes, for matlab;
     * specifically, replace each occurrence of a single quote with a pair of
     * single quotes
     */
    protected static String escapeStringForMatlab(String string)
    {
        return string.replaceAll("'", "''");
    }


    /*
     * find whatever MATLAB installations we can without help from the
     * user. this process is system-dependent; on windows, the registry is
     * used, while (currently) on linux and macos x, only the MATLAB
     * installation indicated by the PATH env. var. is reported.
     *
     * FIXME: rename this
     */
    public abstract Map<File, InstallerGUI.MatlabInstallation> findMatlabPathsAndVersions();


    /*
     *
     */
    protected abstract boolean
        gatherPlatformDependentConfigurationInformation(List<String> errors); // out


    /*
     * gather and store various pieces of system config. info. that may
     * indicate an inability to proceed with installation.
     *
     * return true if all system config. information looks OK;
     * return false otherwise.
     *
     * if false is returned, the errors argument will be filled with one
     * String -- containing human-readable information -- per error condition
     *
     * FIXME: these data should be extracted out into a separate object
     * FIXME: need to re-classify error as fatal or non-fatal, and simply throw an exception
     *        when fatal errors occur
     */
    public boolean gatherSystemConfigurationInformation(List<String> errors) // out
    {
        boolean success = true;
        List<String> platformDependentErrors;

        
        //
        if (getUserHomeDirectory() == null)
        {
            errors.add("Unable to determine home directory");
            success = false;
        }

        //
        platformDependentErrors = new LinkedList<String>();
        success =
            success && gatherPlatformDependentConfigurationInformation(platformDependentErrors);
        
        //
        errors.addAll(platformDependentErrors);
        return success;
    }


    /*
     * given the matlab path, get the abstract path of the matlab
     * binary executable. this may be system-dependent.
     *
     * FIXME: not thrilled about needing two File object allocs.
     */
    protected File getBinaryFilePathFromMatlabPath(File matlabPath)
    {
        return new File(new File(matlabPath, "bin"), "matlab");
    }


    /*
     * gets the value returned by either getConfigDirectory() or
     * getTempConfigDirectory(), depending on whether the temp. dir.
     * has been moved to the final dir. location -- this is just
     * a convenience method
     *
     * should only be called after gatherSystemConfigurationInformation()
     */
    protected File getActiveConfigDirectory()
    {
        if (this.tempConfigDirHasBeenFinalized)
        {
            return getConfigDirectory();
        }
        else
        {
            return getTempConfigDirectory();
        }
    }


    /*
     * <home dir> + "/.wfdb-swig-matlab"
     *
     * should only be called after gatherSystemConfigurationInformation()
     */
    protected File getConfigDirectory()
    {
        return new File(getUserHomeDirectory(), ".wfdb-swig-matlab");
    }


    /*
     * FIXME: not crazy about using three allocs. here!!!
     */
    protected File getMatlabClassPathFile(File matlabPath)
    {
        return new File(new File(new File(matlabPath, "toolbox"), "local"), "classpath.txt");
    }


    /*
     * FIXME: not crazy about using three allocs. here!!!
     */
    protected File getMatlabLibraryPathFile(File matlabPath)
    {
        return new File(new File(new File(matlabPath, "toolbox"), "local"), "librarypath.txt");
    }


    /*
     * if matlab returns an empty userpath string, this is used to make a guess
     * of what it should be; on *nix, it should probably be ~/matlab, and on
     * windows probably ~/My Documents/MATLAB
     */
    protected abstract File getMatlabUserpathGuess();


    /*
     *
     */
    protected abstract Set<String> getNativeBinaryFileNames(String arch)
        throws IOException;


    /*
     * textFileName is the file name of a resource that has been included with the jar
     */
    protected static Set<String> getResourceNamesFromTextFileResource(String textFileName)
        throws IOException
    {
        URL textFileURL;
        InputStream textFileStream;
        StreamTokenizer textFileTokenizer;
        Set<String> resourceNames;


        textFileURL = Thread.currentThread().getContextClassLoader().getResource(textFileName);
        textFileStream = textFileURL.openStream();
        textFileTokenizer =
            new StreamTokenizer(new BufferedReader(new InputStreamReader(textFileStream)));

        // FIXME: this seems a little silly...
        textFileTokenizer.slashSlashComments(false);
        textFileTokenizer.slashStarComments(false);
        textFileTokenizer.wordChars((int)'/', (int)'/');
        textFileTokenizer.wordChars((int)'.', (int)'.');
        textFileTokenizer.wordChars((int)'_', (int)'_');

        resourceNames = new HashSet<String>();
        while (textFileTokenizer.nextToken() != StreamTokenizer.TT_EOF)
        {
            resourceNames.add(textFileTokenizer.sval);
        }

        return resourceNames;
    }


    /*
     *
     */
    protected Password getSuperuserPassword()
    {
        return this.superuserPassword;
    }


    /*
     * <home dir> + "/.wfdb-swig-matlab-tmp"
     *
     * should only be called after gatherSystemConfigurationInformation()
     */
    protected File getTempConfigDirectory()
    {
        return new File(getUserHomeDirectory(), ".wfdb-swig-matlab-tmp");
    }


    /*
     *
     */
    protected File getUserHomeDirectory()
    {
        if (!this.userHomeDirectoryGotten)
        {
            String userHomeDirectoryString;

            userHomeDirectoryString = System.getProperty("user.home");
            if (userHomeDirectoryString == null)
            {
                this.userHomeDirectory = null;
            }
            else
            {
                this.userHomeDirectory = new File(userHomeDirectoryString);
            }
            this.userHomeDirectoryGotten = true;
        }

        return this.userHomeDirectory;
    }


    /*
     *
     */
    public File getWSMCodeDirectory()
    {
        return new File(getConfigDirectory(), convertVersionToString(currentVersion));
    }


    /*
     *
     */
    public File getWSMCodeUtilDirectory()
    {
        return new File(getWSMCodeDirectory(), "util");
    }


    /*
     * get the names of the matlab code files that are packaged for us;
     * this is slightly tricky, because we have to read from a text file
     * (whose name we know) to get the list of matlab code files.
     *
     * FIXME: yes, i would like a way to do this without needing this
     *        auxiliary text file, but i don't know whether such a way
     *        exists.
     */
    private static Set<String> getWSMMatlabCodeFileNames()
        throws IOException
    {
        return getResourceNamesFromTextFileResource(matlabCodeListFileName);
    }


    /*
     *
     */
    protected abstract boolean isFileReadOnly(File file)
        throws IOException,
               ProcessExecutionException;


    /*
     *
     */
    protected abstract boolean isSupportedArch(String arch);


    /*
     *
     */
    protected abstract boolean isSupportedMatlabVersion(String version);


    /*
     * launch matlab graphically, with the given command, and do not
     * wait for it to return.
     *
     * matlabCommand must not contain any semi-colons or
     * whitespace (most likely it will be just a single
     * function call without arguments)
     *
     * this is used, e.g., to launch matlab with the demo script just as
     * the installer is quitting.
     *
     * FIXME: actually, the installer should not quit (but also should not
     *        have a timeout), because it may have to do cleanup still, in
     *        case the windows .bat file is used
     */
    protected abstract void launchMatlabWithCommand(File matlabPath,
                                                    String matlabCommand,
                                                    File workingDirectory)
        throws IOException;


    /*
     * assuming a relative path that uses '/' as the directory-separator
     * character, localize it to the current platform (e.g. on Windowss
     * replace '/' with '\\')
     */
    protected static String localizeRelativePath(String path)
    {
        return path.replace("/", File.separator);
    }


    /*
     * essentially a wrapper around File.mkdir() that throws
     * an IOException upon failure
     */
    private static void makeDirectory(File dirToMake)
        throws IOException
    {
        if (!dirToMake.mkdir())
        {
            throw new IOException("Unable to create directory: " + dirToMake.getAbsolutePath());
        }
    }


    /*
     *
     */
    public void moveTempConfigDirToFinalConfigDir()
        throws IOException
    {
        File tempConfigDir, finalConfigDir;

        
        //
        tempConfigDir = getTempConfigDirectory();
        finalConfigDir = getConfigDirectory();

        //
        if (finalConfigDir.exists())
        {
            removeDirectory(finalConfigDir);
        }

        //
        if (!tempConfigDir.renameTo(finalConfigDir))
        {
            throw new IOException("Unable to rename: "
                                  + tempConfigDir.getAbsolutePath() + " to "
                                  + finalConfigDir.getAbsolutePath());
        }
        else
        {
            this.tempConfigDirHasBeenFinalized = true;
            this.logger.fine("Renamed dir. " + tempConfigDir.getAbsolutePath() + " to "
                             + finalConfigDir.getAbsolutePath() + "\n");
        }
    }


    /*
     * FIXME: is this really not covered by anything in the java API???
     * FIXME: move this out to some sort of "Util" class
     */
    public static void removeDirectory(File dirToDelete)
        throws IOException
    {
        Stack<File> filesToDelete;
        Logger logger = Logger.getLogger("controller");   // FIXME: a bit improper


        //
        assert (dirToDelete != null);

        //
        logger.info("About to remove directory " + dirToDelete.getAbsolutePath() + " "
                    + "and all of its contents ...\n");

        //
        filesToDelete = new Stack<File>();
        filesToDelete.push(dirToDelete);
        while (!filesToDelete.empty())
        {
            File file = filesToDelete.pop();

            if (file.isDirectory())
            {
                File[] dirFiles = file.listFiles();

                if (dirFiles == null)
                {
                    throw new IOException("Unable to list files in directory: "
                                          + file.getAbsolutePath());
                }
                else if (dirFiles.length == 0)
                {
                    // try to remove the empty directory
                    deleteFile(file);
                }
                else
                {
                    // push back the directory, and everything it contains
                    filesToDelete.push(file);
                    for (File f : dirFiles)
                    {
                        filesToDelete.push(f);
                    }
                }
            }
            else
            {
                // try to remove the ordinary (non-directory) file
                deleteFile(file);
            }
        }

        //
        logger.info("Finished removing directory " + dirToDelete.getAbsolutePath() + ".\n");
    }


    /*
     * the default version of runMatlabCommand(): just capture the last
     * line of the output as a String
     */
    protected String runMatlabCommand(File matlabPath,
                                      String matlabCommand,
                                      File logFile)
        throws IOException,
               ProcessExecutionException
    {
        return runMatlabCommand(matlabPath,
                                matlabCommand,
                                logFile,
                                new ProcessOutputAnalyzer.LastLine());
    }


    /*
     * a number of methods (including some public ones) can be written
     * in terms of this system-dependent method.
     *
     * all arguments except logFile must be non-null
     *
     * note that even if logFile is null, a log file may be created, as
     * e.g. it is the only reasonable way to get output on windows
     * (of which i know).
     */
    protected abstract <ResultType> ResultType
        runMatlabCommand(File matlabPath,
                         String matlabCommand,
                         File logFile,
                         ProcessOutputAnalyzer<ResultType> outputAnalyzer)
        throws IOException,
               ProcessExecutionException;


    /*
     * pathIsCorrect is set to true if test_wfdb appears to be in the path
     * by default, as we want it to be; false otherwise
     *
     * wfdbTestWasSuccessful is set to true if test_wfdb succeeded;
     * false otherwise
     *
     * an Exception may be thrown if matlab was executed properly,
     * but an error otherwise occurred
     */
    public void runWFDBTest(File matlabPath,                         // in
                            Pointer<Boolean> pathIsCorrect,          // out
                            Pointer<Boolean> wfdbTestIsSuccessful)   // out
        throws IOException,
               ProcessExecutionException
    {
        File wsmCodeDir, wsmCodeUtilDir;
        String matlabCommand;
        String result;


        // FIXME: these are redundant with calls made elsewhere
        //        (e.g. in copyFilesFromPackageToHost())
        wsmCodeDir = new File(getConfigDirectory(), convertVersionToString(currentVersion));
        wsmCodeUtilDir = new File(wsmCodeDir, "util");

        //
        matlabCommand =
            "try;path(path,pathdef);catch;end;"
            + "try;"
            + "p=1;"
            + "t=test_wfdb;"
            + "catch;"
            + "p=0;"
            + "addpath('" + escapeStringForMatlab(wsmCodeDir.getAbsolutePath()) + "');"
            + "addpath('" + escapeStringForMatlab(wsmCodeUtilDir.getAbsolutePath()) + "');"
            + "t=test_wfdb;"
            + "end;"
            + "nl=['\\','n'];"
            + "fprintf(1,['%d%d',nl],p,t);"
            + "exit;";

        //
        result = runMatlabCommand(matlabPath, matlabCommand, null);
        result = result.trim();

        // verify that the correct values were returned by matlab
        pathIsCorrect.object = result.substring(0, 1).equals("1");
        wfdbTestIsSuccessful.object = result.substring(1, 2).equals("1");
    }


    /*
     *
     */
    protected abstract void setFileReadOnly(File file,
                                            boolean readOnly)
        throws IOException,
               ProcessExecutionException;


    /*
     *
     */
    public void setGUI(InstallerGUI gui)
    {
        this.gui = gui;
    }


    // /*
    //  *
    //  */
    // public abstract void setLogger(Logger logger);


    /*
     * FIXME: do not use localizeRelativePath()!
     */
    private void updateMatlabClassPathFileToTempFile(File classPathFile,
                                                     WSMInstallationData newWSMInstall,
                                                     File tempClassPathFile)
        throws FileNotFoundException,
               IOException
    {
        LineFilter filter;
        String newCommentText;
        String[] newClassPathStrings;
        File wsmInstallDir; // String wsmInstallDir;
        String wfdbJarPath;
        String wsmClassesPath;


        //
        filter = new MatlabPathFileWSMLineFilter();

        // FIXME: this is an annoying hack; the full install dir. path
        //        should be encapsulated elsewhere
        // wsmInstallDir = getConfigDirectory().getAbsolutePath()
        //                 + "/" + convertVersionToString(newWSMInstall.getWSMVersion());
        wsmInstallDir = new File(getConfigDirectory(),
                                 convertVersionToString(newWSMInstall.getWSMVersion()));

        //
        wfdbJarPath =
            escapeAbsolutePathForPathFile((new File(wsmInstallDir,
                                                    "wfdb.jar")).getAbsolutePath());
        wsmClassesPath =
            escapeAbsolutePathForPathFile((new File(wsmInstallDir,
                                                    "wsm-classes.jar")).getAbsolutePath());

        //
        newClassPathStrings = new String[4];

        newCommentText = String.format(pathFileCommentText,
                                       convertVersionToString(newWSMInstall.getWSMVersion()),
                                       newWSMInstall.getUUID(),
                                       wfdbJarPath);
                                       // localizeRelativePath(wsmInstallDir + "/wfdb.jar"));
        newClassPathStrings[0] = "#" + newCommentText;
        newClassPathStrings[1] = wfdbJarPath; // localizeRelativePath(wsmInstallDir + "/wfdb.jar");

        newCommentText = String.format(pathFileCommentText,
                                       convertVersionToString(newWSMInstall.getWSMVersion()),
                                       newWSMInstall.getUUID(),
                                       wsmClassesPath);
                                       // localizeRelativePath(wsmInstallDir + "/wsm-classes.jar"));
        newClassPathStrings[2] = "#" + newCommentText;
        newClassPathStrings[3] = wsmClassesPath; // localizeRelativePath(wsmInstallDir + "/wsm-classes.jar");

        //
        updateMatlabPathFileToTempFile(classPathFile, filter,
                                       newClassPathStrings, tempClassPathFile);
    }


    /*
     * FIXME: do not use localizeRelativePath()!
     */
    private void updateMatlabLibraryPathFileToTempFile(File libraryPathFile,
                                                       WSMInstallationData newWSMInstall,
                                                       File tempLibraryPathFile,
                                                       String arch)
        throws FileNotFoundException,
               IOException
    {
        LineFilter filter;
        String newCommentText;
        String[] newLibraryPathStrings;
        File wsmInstallDir; // String wsmInstallDir;
        String libPath;


        //
        filter = new MatlabPathFileWSMLineFilter();

        // FIXME: this is an annoying hack; the full install dir. path
        //        should be encapsulated elsewhere
        // wsmInstallDir = getConfigDirectory().getAbsolutePath()
        //                 + "/" + convertVersionToString(newWSMInstall.getWSMVersion());
        wsmInstallDir = new File(getConfigDirectory(),
                                 convertVersionToString(newWSMInstall.getWSMVersion()));

        //
        libPath =
            escapeAbsolutePathForPathFile((new File(wsmInstallDir,
                                                    "lib-" + arch)).getAbsolutePath());

        //
        newLibraryPathStrings = new String[2];
        
        newCommentText = String.format(pathFileCommentText,
                                       convertVersionToString(newWSMInstall.getWSMVersion()),
                                       newWSMInstall.getUUID(),
                                       libPath);
                                       // localizeRelativePath(wsmInstallDir + "/lib-" + arch));
        newLibraryPathStrings[0] = "#" + newCommentText;
        newLibraryPathStrings[1] = libPath; // localizeRelativePath(wsmInstallDir + "/lib-" + arch);

        //
        updateMatlabPathFileToTempFile(libraryPathFile, filter,
                                       newLibraryPathStrings, tempLibraryPathFile);
    }


    /*
     * update librarypath.txt and classpath.txt for the given
     * matlab installation (note the double use of the term "path,"
     * referring generally to these two files, and also -- using
     * the more typical definition -- referring to the location
     * where matlab is installed.)
     *
     * FIXME: setFileReadOnly() is only really relevant to windows --
     *        it should be moved into the windows NativeController.copyFile()
     */
    public void updateMatlabPathFiles(File matlabPath,
                                      String matlabVersion,
                                      String arch)
        throws IOException,
               NoSuperuserPasswordException,
               ProcessExecutionException
    {
        File classPathFile, tempClassPathFile;
        File libraryPathFile, tempLibraryPathFile;
        WSMInstallationData newInstall;


        // FIXME: this should not be created here -- it should be created once
        //        per installation (probably by the GUI)
        newInstall = new WSMInstallationData(currentVersion,
                                             matlabVersion,
                                             matlabPath.getAbsolutePath(),
                                             this.currentUUID);

        //
        classPathFile = getMatlabClassPathFile(matlabPath);
        libraryPathFile = getMatlabLibraryPathFile(matlabPath);

        // FIXME: because of xcopy on windows, these temp. filenames must be the same
        //        as that of their final destinations, e.g. no ".tmp" prefix is allowed!
        tempClassPathFile = new File(getActiveConfigDirectory(), "classpath.txt");
        tempLibraryPathFile = new File(getActiveConfigDirectory(), "librarypath.txt");

        //
        updateMatlabClassPathFileToTempFile(classPathFile, newInstall, tempClassPathFile);
        updateMatlabLibraryPathFileToTempFile(libraryPathFile, newInstall,
                                              tempLibraryPathFile, arch);

        // copy the temp. path files to their final destinations
        for (File[] f : new File[][] { new File[] { tempClassPathFile, classPathFile },
                                       new File[] { tempLibraryPathFile, libraryPathFile } })
        {
            File tempPathFile = f[0];
            File finalPathFile = f[1];
            boolean fileIsReadOnly;
            
            //
            fileIsReadOnly = isFileReadOnly(finalPathFile);
            if (fileIsReadOnly)
            {
                setFileReadOnly(finalPathFile, false);
            }

            //
            try
            {
                copyFile(tempPathFile, finalPathFile);
            }
            finally
            {
                if (fileIsReadOnly)
                {
                    setFileReadOnly(finalPathFile, true);
                }
            }                
        }
    }


    /*
     * read either classpath.txt or librarypath.txt, filtering out the lines to
     * be removed, and then append new lines, writing the result into a temp.
     * file
     */
    private void updateMatlabPathFileToTempFile(File pathFile,
                                                LineFilter filter,
                                                String[] linesToAppend,
                                                File tempPathFile)
        throws FileNotFoundException,
               IOException
    {
        FileWriter fileWriter;
        String outputText; // FIXME: use a StringBuffer
        String newline;


        //
        newline = System.getProperty("line.separator");

        //
        if (pathFile.exists())
        {
            BufferedReader lineReader = new BufferedReader(new FileReader(pathFile));

            outputText = "";
            try
            {
                String line;
                
                while ((line = lineReader.readLine()) != null)
                {
                    if (filter.matches(line))
                    {
                        outputText += line + newline;
                    }
                }
            }
            finally   // propagate up any uncaught IOException
            {
                try
                {
                    lineReader.close();
                }
                catch (IOException ex)
                {
                    this.logger.info("Unable to close " + pathFile.getAbsolutePath() + " "
                                     + "properly: " + ex.getMessage() + " (non-fatal)\n");
                }
            }
        }
        else
        {
            outputText = "";
        }

        //
        for (String line : linesToAppend)
        {
            outputText += line + newline;
        }

        //
        fileWriter = new FileWriter(tempPathFile);
        try
        {
            fileWriter.write(outputText);
        }
        finally   // propagate up any uncaught IOException
        {
            try
            {
                fileWriter.close();
            }
            catch (IOException ex)
            {
                this.logger.warning("Unable to close " + tempPathFile.getAbsolutePath() + " "
                                    + "properly: " + ex.getMessage() + " (may be problematic)\n");
            }
        }
    }


    /*
     * this is a difficult method to get right, because of matlab's less
     * than spectacular path handling. it tries to use savepath(), and will
     * revert to path2rc() in case of error. if the matlab userpath variable was
     * non-empty, this will call savepath() (or path2rc()) both without an argument,
     * and with userpath+"/pathdef.m".
     *
     * FIXME: the trouble is that this may update the system-wide pathdef.m (in
     *        toolbox/local/) when it doesn't make sense to do so; not a big problem,
     *        and more robust for WSM to work, but a bit of an overreach.
     *
     * FIXME: if matlab claimed that userpath was empty, it might have also been the
     *        case that it *wouldn't* have been empty if the right directory existed
     *        in the filesystem. i.e., if ~/matlab exists, userpath is "~/matlab:",
     *        but if it doesn't, userpath is "". we might want to guess the userpath,
     *        in the latter case.
     *
     * FIXME: this would be faster with StringBuffer
     */
    public void updateMatlabPathVariable(File matlabPath,
                                         InstallerGUI.MatlabInstallation matlabInstall)
        throws IOException,
               ProcessExecutionException
    {
        File wsmCodeDir, wsmCodeUtilDir;
        String matlabCommand;
        File userpathToTry;


        // FIXME: these are redundant with calls made elsewhere
        //        (e.g. in copyFilesFromPackageToHost())
        wsmCodeDir = new File(getConfigDirectory(), convertVersionToString(currentVersion));
        wsmCodeUtilDir = new File(wsmCodeDir, "util");

        //
        if (matlabInstall.primaryUserpath != null)
        {
            userpathToTry = matlabInstall.primaryUserpath;

            this.logger.info("MATLAB version " + matlabInstall.fullVersion + " "
                             + "gave userpath " + userpathToTry.getAbsolutePath() + "\n");
        }
        else
        {
            userpathToTry = getMatlabUserpathGuess();

            this.logger.info("MATLAB version " + matlabInstall.fullVersion + " "
                             + "did not provide a userpath" +
                             ((userpathToTry != null)
                              ? ("; guessing " + userpathToTry.getAbsolutePath())
                              : "")
                             + "\n");
        }

        //
        matlabCommand =
            "addpath('" + escapeStringForMatlab(wsmCodeDir.getAbsolutePath()) + "');"
            + "addpath('" + escapeStringForMatlab(wsmCodeUtilDir.getAbsolutePath()) + "');";

        //if (matlabInstall.primaryUserpath != null)
        if (userpathToTry != null)
        {
            String userpathParent = userpathToTry.getParent(); // matlabInstall.primaryUserpath.getParent();
            String userpathDirName = userpathToTry.getName(); // matlabInstall.primaryUserpath.getName();
            File userpathPathdefFile = new File(userpathToTry, "pathdef.m");
                //new File(matlabInstall.primaryUserpath, "pathdef.m");

            if (userpathParent != null && userpathParent.length() > 0
                && userpathDirName != null && userpathDirName.length() > 0)
            {
                matlabCommand +=
                    "mkdir('" + escapeStringForMatlab(userpathParent)
                    + "','" + escapeStringForMatlab(userpathDirName) + "');";
            }

            matlabCommand +=
                "try;"
                + "savepath;"
                + "savepath('" + escapeStringForMatlab(userpathPathdefFile.getAbsolutePath()) + "');"
                + "catch;"
                + "path2rc;"
                + "path2rc('" + escapeStringForMatlab(userpathPathdefFile.getAbsolutePath()) + "');"
                + "end;"
                + "exit;";
        }
        else
        {
            matlabCommand +=
                "try;"
                + "savepath;"
                + "catch;"
                + "path2rc;"
                + "end;"
                + "exit;";
        }

        //
        runMatlabCommand(matlabPath, matlabCommand, null, null);
    }


    /*
     * given a potential matlab path, verify that it's valid (by attempting to run it),
     * and return:
     *   - the userpath
     *   - the full matlab version
     *   - the JVM arch.
     *
     * in case of failure, log the error and return null.
     *
     * FIXME: put try/catch around the java os.arch call
     * FIXME: rename this
     */
    public InstallerGUI.MatlabInstallation verifyMatlabPathAndGetVersion(File matlabPath)
    {
        final String matlabVersionAndArchCommand =
            "nl=['\\','n'];fprintf(1,['%s',nl,'%s',nl,'%s',nl], "
            + "userpath, "
            + "version, "
            + "char(java.lang.System.getProperty('os.arch')));"
            + "exit;";

        //
        try
        {
            String fullUserpath, primaryUserpath;
            String arch;
            String version;
            List<String> fullOutput;

            //
            fullOutput = runMatlabCommand(matlabPath, matlabVersionAndArchCommand,
                                          null, new ProcessOutputAnalyzer.EveryLine());
            if (fullOutput == null || fullOutput.size() < 2)
            {
                this.logger.severe("Unable to get version and/or arch. for MATLAB in "
                                   + matlabPath.getAbsolutePath() + "\n");
                return null;
            }

            //
            fullUserpath = fullOutput.get(fullOutput.size() - 3);
            version = fullOutput.get(fullOutput.size() - 2);
            arch = fullOutput.get(fullOutput.size() - 1);

            //
            if (fullUserpath == null)
            {
                fullUserpath = "";
            }

            if (fullUserpath.length() == 0)
            {
                primaryUserpath = "";
            }
            else
            {
                String[] userpathDirs;
                int i;

                fullUserpath = fullUserpath.trim();
                // FIXME: this assumes that File.pathSeparator contains only plain characters
                //        (i.e. nothing interpreted specially in a regex)
                // FIXME: this also assumes that the directory names in the path cannot include
                //        File.pathSeparator; this appears to be true, but i haven't found it
                //        documented
                userpathDirs = fullUserpath.split(File.pathSeparator);
                for (i = 0, primaryUserpath = ""; i < userpathDirs.length; ++i)
                {
                    if (userpathDirs[i] != null && userpathDirs[i].length() > 0)
                    {
                        primaryUserpath = userpathDirs[i];
                        break;
                    }
                }
            }

            //
            if (primaryUserpath.length() == 0)
            {
                this.logger.warning("MATLAB in " + matlabPath.getAbsolutePath() + " has no userpath; "
                                    + "that's weird.\n");
            }

            //
            if (arch == null || !isSupportedArch(arch))
            {
                this.logger.warning("Unable to use MATLAB in " + matlabPath.getAbsolutePath() + " -- "
                                    + "the JVM's architecture is "
                                    + ((arch == null)
                                       ? "unknown (Java property os.arch missing!)"
                                       : (arch + ", which is currently unsupported."))
                                    + "\n");
                return null;
            }

            //
            if (version == null || !isSupportedMatlabVersion(version))
            {
                this.logger.warning("Unable to use MATLAB in " + matlabPath.getAbsolutePath() + " -- "
                                    + "version "
                                    + ((version == null)
                                       ? "could not be determined."
                                       : (" is " + version + ", which is currently unsupported."))
                                    + "\n");
                return null;
            }

            //
            return new InstallerGUI.MatlabInstallation(version,
                                                       arch,
                                                       (primaryUserpath.length() > 0)
                                                       ? new File(primaryUserpath)
                                                       : null);
        }
        catch (IOException ex)
        {
            this.logger.warning("Unable to verify MATLAB installation in " + matlabPath + ": "
                                + ex.getMessage() + "\n");
            return null;            
        }
        catch (ProcessExecutionException ex)
        {
            this.logger.warning("Unable to verify MATLAB installation in " + matlabPath + ": "
                                + ex.getMessage() + "\n");
            return null;
        }
    }


    /*
     * return true if the superuser password is correct; false if not
     */
    protected abstract boolean verifySuperuserPassword(Password superuserPassword);


    /*
     *
     */
    public void writeConfigFile(File matlabPath,
                                String matlabVersion)
        throws IOException
    {
        File configFile;
        FileWriter writer;

        
        //
        configFile = new File(getActiveConfigDirectory(), matlabVersion + ".conf");
        if (configFile.exists())
        {
            deleteFile(configFile);
        }

        writer = new FileWriter(configFile);
        writer.write(this.currentUUID + "\n");
        writer.write(convertVersionToString(currentVersion) + "\n");
        writer.write(matlabVersion + "\n");
        writer.write(matlabPath.getAbsolutePath() + "\n");
        writer.close();
    }
}
