/*
 *
 */


/*
 *
 */
package installer.common;


/*
 *
 */
import java.awt.EventQueue;
import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;
import javax.swing.JOptionPane;


/*
 *
 */
public class InstallerGUI
{
    /*
     *
     */
    private InstallerWindow installerWindow;

    /*
     *
     */
    private Controller nativeController;

    /*
     *
     */
    private NativeGUI nativeGUI;


    /*
     * when a MATLAB installation is indicated for installation,
     * it goes through a few steps:
     *
     * - verification:
     *     a check to make sure it really exists, can be executed,
     *     and (in the future) has a current-enough version
     * - installation:
     *     updating librarypath.txt and classpath.txt files, and writing
     *     the appropriate config. file to ~/.wfdb-swig-matlab-tmp
     * - testing:
     *     running the WFDB test script to verify that it works properly
     *
     * we keep track of all the known MATLAB installations, how far they
     * have progressed through these steps, and whether or not they have
     * failed (if they fail any step, they cannot continue through the
     * rest.)
     */


    /*
     *
     */
    public static enum InstallationStep
    {        
        NONE,
        UNSELECTED,
        VERIFICATION,
        INSTALLATION,
        TESTING;
    }


    /*
     * this is essentially used as a struct;
     */
    public static class MatlabInstallation
    {
        //
        public final String fullVersion;
        public final String jvmArch;
        public final File primaryUserpath;
        public InstallationStep lastFinishedStep;
        public boolean hasFailed;

        // FIXME: this is a bit of a hack; it is false if the call to
        //        updateMatlabPathVariable() failed, and true otherwise,
        //        *until* we test it later, at which point it might go false again.
        //        (even if updateMatlabPathVariable() appears to have run
        //        successfully, this is not a guarantee that the path was
        //        updated properly.)
        public boolean pathUpdated;

        //
        public MatlabInstallation(String fullVersion,
                                  String jvmArch,
                                  File primaryUserpath)
        {
            this.fullVersion = fullVersion;
            this.jvmArch = jvmArch;
            this.primaryUserpath = primaryUserpath;
            this.lastFinishedStep = InstallationStep.NONE;
            this.hasFailed = false;
            this.pathUpdated = false;
        }
    }


    /*
     *
     */
    private Map<File, MatlabInstallation> matlabInstallations;


    /*
     * FIXMEFIXME: the following maps are deprecated!!!
     */

    // /*
    //  * a map of MATLAB paths to versions:
    //  * passed verification, but failed installation
    //  */
    // private Map<File, String> failedInstallationMatlabs;


    // /*
    //  * a map of MATLAB paths to versions:
    //  * passed verification, and passed installation, but failed testing
    //  */
    // private Map<File, String> failedTestingMatlabs;
    
    // /*
    //  * a map of MATLAB paths to versions:
    //  * failed verification
    //  */
    // private Map<File, String> failedVerificationMatlabs;


    // /*
    //  * a map of MATLAB paths to versions:
    //  * passed all steps
    //  */
    // private Map<File, String> succeededMatlabs;


    // /*
    //  * a map of MATLAB paths to versions:
    //  * passed verification, but not installed-for, and not tested
    //  */
    // private Map<File, String> uninstalledMatlabs;


    // /*
    //  * a map of MATLAB paths to versions:
    //  * passed verification, passed installation, but not tested
    //  */
    // private Map<File, String> untestedMatlabs;


    /*
     *
     */
    public InstallerGUI(NativeGUI nativeGUI,
                        final Logger logger,
                        final Controller nativeController)
    {
        final InstallerGUI guiThis = this;


        //
        this.installerWindow = null;
        this.nativeController = nativeController;
        this.nativeGUI = nativeGUI;

        //
        this.matlabInstallations = new HashMap<File, MatlabInstallation>();

        // //
        // this.uninstalledMatlabs = new HashMap<File, String>();
        // this.untestedMatlabs = new HashMap<File, String>();

        //
        this.nativeController.setGUI(this);

        //
        EventQueue.invokeLater(new Runnable()
            {
                public void run()
                {
                    installerWindow = new InstallerWindow("WFDB Tools for MATLAB",
                                                          guiThis,
                                                          nativeController,
                                                          logger);
                    installerWindow.show();
                }
            });
    }


    // /*
    //  * move the given MATLAB install. from uninstalledMatlabs -> untestedMatlabs
    //  *
    //  * if the given path and version are not in the set of uninstalledMatlabs,
    //  * then silently do nothing.
    //  */
    // protected void addInstalledMatlab(File matlabPath,
    //                                   String matlabVersion)
    // {
    //     String matlabVersionFromMap;


    //     //
    //     matlabVersionFromMap = this.uninstalledMatlabs.remove(matlabPath);
    //     if (!matlabVersion.equals(matlabVersionFromMap))
    //     {
    //         // FIXME: this is a bug -- log it, if it occurs
    //         return;
    //     }

    //     this.untestedMatlabs.put(matlabPath, matlabVersion);
    // }


    /*
     * FIXME: slightly inefficient to make it newly unmodifiable each time...
     */
    protected Map<File, MatlabInstallation> getMatlabInstallations()
    {
        return Collections.unmodifiableMap(this.matlabInstallations);
    }


    // /*
    //  * FIXME: slightly inefficient to make it newly unmodifiable each time...
    //  */
    // protected Map<File, String> getUninstalledMatlabs()
    // {
    //     return Collections.unmodifiableMap(this.uninstalledMatlabs);
    // }


    // /*
    //  * FIXME: slightly inefficient to make it newly unmodifiable each time...
    //  */
    // protected Map<File, String> getUntestedMatlabs()
    // {
    //     return Collections.unmodifiableMap(this.untestedMatlabs);
    // }


    /*
     * record this information about a MATLAB installation at the given
     * path, with the given version, that has been given a preliminary
     * verification (enough, e.g., to execute and get its version), but
     * for which wfdb-swig-matlab has not yet been fully installed
     * (e.g. its path files not updated, and the test script not executed
     * by it)
     *
     * FIXME: rename this
     */
    protected void giveMatlabPathAndVersion(File path,
                                            MatlabInstallation matlabInst)
    {
        // MatlabInstallation newInst;


        // newInst = new MatlabInstallation(fullVersion);
        // newInst.lastFinishedStep = InstallationStep.VERIFICATION;
        // this.matlabInstallations.put(path, newInst);
        this.matlabInstallations.put(path, matlabInst);
    }

    
    /*
     *
     */
    public void showRootPasswordIncorrectDialog()
    {
        try
        {
            EventQueue.invokeAndWait(new Runnable()
                {
                    public void run()
                    {
                        // FIXME: would like to be able to provide the panel in the first arg.
                        JOptionPane.showMessageDialog(null,
                                                      "The password provided is incorrect.",
                                                      "Incorrect Password",
                                                      JOptionPane.ERROR_MESSAGE);
                    }
                });
        }
        catch (InterruptedException ex)
        {
            // FIXME: this should never happen, but should probably log it...
        }
        catch (InvocationTargetException ex)
        {
            // FIXME: this should never happen, but should probably log it...                
        }
    }


    /*
     *
     */
    public Controller.Password showRootPasswordPrompt()
    {
        final Pointer<Controller.Password> password = new Pointer<Controller.Password>(null);


        //
        try
        {
            EventQueue.invokeAndWait(new Runnable()
                {
                    public void run()
                    {
                        password.object = nativeGUI.showRootPasswordPrompt();
                    }
                });
        }
        catch (InterruptedException ex)
        {
            // FIXME: this should never happen, but should probably log it...
        }
        catch (InvocationTargetException ex)
        {
            // FIXME: this should never happen, but should probably log it...                
        }
        
        return password.object;
    }
}

