/*
 *
 */


/*
 *
 */
package installer.common;


/*
 *
 */
import java.io.BufferedReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


/*
 * interface for a (typically line-based) analyzer of the output
 * of a process. the interface is fairly simple; most of the value
 * comes from the few implementing classes, provided here, that
 * cover most/much of the usage.
 *
 * FIXME: should this be templated <T>, with a public T getResult() method?
 * FIXME: +--> YES! use exceptions with detailed messages about what went
 *             wrong, and return the return values from analyzeOutput()
 */
public interface ProcessOutputAnalyzer<ResultType>
{
    /*
     * return true if the output provided matched the form expected,
     * and so was analyzed properly;
     * return false otherwise.
     */
    public ResultType analyzeOutput(BufferedReader outputReader)
        throws IOException,
               UnexpectedOutputException;


    /*
     * thrown by ProcessOutputAnalyzer.analyzeOutput() when the
     * output did not meet expectations
     */
    public class UnexpectedOutputException extends Exception
    {
        /*
         *
         */
        public UnexpectedOutputException(String message)
        {
            super(message);
        }
    }


    /*
     * get the very last line of the output
     */
    public static class LastLine implements ProcessOutputAnalyzer<String>
    {
        /*
         *
         */
        public String analyzeOutput(BufferedReader outputReader)
            throws IOException,
                   UnexpectedOutputException
        {
            String line;
            String savedLine;


            savedLine = null;
            while ((line = outputReader.readLine()) != null)
            {
                savedLine = line;
            }

            if (savedLine != null)
            {
                return savedLine;
            }
            else
            {
                throw new UnexpectedOutputException("No output produced");
            }
        }
    }


    /*
     * try to match every line of the output against a given regular expression,
     * which should have a single group, and store the set of those matched lines'
     * Strings captured by the group.
     */
    public static class SingleGroupRegexpAll implements ProcessOutputAnalyzer<List<String>>
    {
        /*
         *
         */
        private String regexpPatternString;

        /*
         * patternString must be non-null
         */
        public SingleGroupRegexpAll(String patternString)
        {
            assert (patternString != null);

            this.regexpPatternString = patternString;
        }

        /*
         *
         */
        public List<String> analyzeOutput(BufferedReader outputReader)
            throws IOException,
                   UnexpectedOutputException
        {
            String line;
            List<String> matchedStrings;
            Pattern regexpPattern;

            
            regexpPattern = Pattern.compile(this.regexpPatternString);
            matchedStrings = new ArrayList<String>();
            while ((line = outputReader.readLine()) != null)
            {
                Matcher regexpMatch = regexpPattern.matcher(line);

                if (regexpMatch.matches())
                {
                    matchedStrings.add(regexpMatch.group(1));
                }
            }

            return Collections.unmodifiableList(matchedStrings);
        }
    }


    /*
     * like SingleGroupRegexpAll, but stops just after finding the first match
     */
    public static class SingleGroupRegexpFirst implements ProcessOutputAnalyzer<String>
    {
        /*
         *
         */
        private String regexpPatternString;

        /*
         * patternString must be non-null
         */
        public SingleGroupRegexpFirst(String patternString)
        {
            assert (patternString != null);

            this.regexpPatternString = patternString;
        }

        /*
         * return true if and only if the pattern ever matched
         */
        public String analyzeOutput(BufferedReader outputReader)
            throws IOException,
                   UnexpectedOutputException
        {
            String line;
            String matchedString;
            Pattern regexpPattern;


            regexpPattern = Pattern.compile(this.regexpPatternString);
            matchedString = null;
            while ((line = outputReader.readLine()) != null)
            {
                Matcher regexpMatch = regexpPattern.matcher(line);
                
                if (regexpMatch.matches())
                {
                    matchedString = regexpMatch.group(1);
                    break;
                }
            }

            if (matchedString != null)
            {
                return matchedString;
            }
            else
            {
                throw new UnexpectedOutputException("No line matching reg. exp.: " + this.regexpPatternString);
            }
        }
    }


    /*
     * this is essentially a decorator.
     *
     * NOTE (FIXME?): this only works if the enclosed analyzer uses
     *                only readLine() on its given BufferedReader
     */
    public static class FirstLineOnly<ResultType> implements ProcessOutputAnalyzer<ResultType>
    {
        /*
         *
         */
        private ProcessOutputAnalyzer<ResultType> decoratedAnalyzer;

        /*
         * analyzer should be non-null
         */
        public FirstLineOnly(ProcessOutputAnalyzer<ResultType> analyzer)
        {
            assert (analyzer != null);

            this.decoratedAnalyzer = analyzer;
        }

        /*
         * return what the decorated analyzer returns
         */
        public ResultType analyzeOutput(BufferedReader outputReader)
            throws IOException,
                   UnexpectedOutputException
        {
            return decoratedAnalyzer.analyzeOutput(new BufferedReader(outputReader) {
                    
                    private boolean readLineCalledYet = false;

                    public String readLine()
                        throws IOException
                    {
                        if (!this.readLineCalledYet)
                        {
                            this.readLineCalledYet = true;
                            return super.readLine();
                        }
                        else
                        {
                            return null;
                        }
                    }
                });
        }
    }


    /*
     * simply return every line of output.
     *
     * this is a special case of SingleGroupRegexpAll, so we simply
     * inherit from it
     */
    public static class EveryLine extends SingleGroupRegexpAll
    {
        /*
         *
         */
        public EveryLine()
        {
            super("(.*)");
        }
    }
}

